from scipy.optimize import _cobyla, fmin_slsqp
# from openopt import NLP

from scipy.special import erf, erfinv, gammainc, gammaincinv
from math import sqrt, isnan
import numpy
import functools
from model import _dict_to_vector, _vector_to_dict, _dict_to_matrix, \
    ParameterError

# import nlopt
# import numdifftools

class FitError(Exception):
    """
    Exception indicating a critical failure during an optimization.

    Args:
      message (str): Error message
      model (Model instance): Model object attempted to fit.
      kwargs (dict): Keyword arguments passed to the ``fit`` function.

    """
    def __init__(self, message, model, kwargs):
        Exception.__init__(self, message, model, kwargs)
        self.model = model
        self.kwargs = kwargs

@functools.total_ordering
class FitResult:
    """
    Stores the results of a ``localFit`` call.

    Note:
      For each keyword argument there is an attribute with the same name.
      All keyword arguments default to ``None``.

    Args:
      bestFitParameters (dict, optional): Holds the parameter values at the
        best-fit point.
      minChiSquare (float, optional): The chi-square value at the minimum.
      success (bool, optional): ``True`` if the minimization terminated
        successfully.
      feasible (bool, optional): ``True`` if the solution has an acceptably small
        constraint violation.
      status (int, optional): Exit code returned by the minimizer.
      message (str, optional): Message returned by the minimizer.
      nEvaluations (int, optional): Number of function evaluations.
      constraintViolation (float, optional): Sum of the squares of the constraint
        functions which violate constraints.

    """
    def __init__(self,
                 bestFitParameters=None,
                 minChiSquare=None, 
                 status=None,
                 success=None,
                 feasible=None,
                 message=None,
                 nEvaluations=None,
                 constraintViolation=None):
        self.bestFitParameters = bestFitParameters
        self.minChiSquare = minChiSquare
        self.status = status
        self.success = success
        self.feasible = feasible
        self.message = message
        self.nEvaluations = nEvaluations
        self.constraintViolation = constraintViolation

    def __repr__(self):
        return 'FitResult('+\
            'bestFitParameters='+repr(self.bestFitParameters)+', '+\
            'minChiSquare='+repr(self.minChiSquare)+', '+\
            'success='+repr(self.success)+', '+\
            'feasible='+repr(self.feasible)+', '+\
            'status='+repr(self.status)+', '+\
            'message='+repr(self.message)+', '+\
            'nEvaluations='+repr(self.nEvaluations)+', '+\
            'constraintViolation='+repr(self.constraintViolation)+')'

    def __str__(self):
        res = 'best-fit parameters:\n'
        keys = sorted(self.bestFitParameters.keys())
        width = max(len(k) for k in keys)
        for k in sorted(self.bestFitParameters.keys()):
            res = res+'  '+k.ljust(width, ' ')+' = '+\
                str(self.bestFitParameters[k])+'\n'
        res = res + 'minimum chi-square:    ' + \
            str(self.minChiSquare) + '\n'
        res = res + 'number of evaluations: ' + \
            str(self.nEvaluations) + '\n'
        res = res + 'constraint violation:  ' + \
            str(self.constraintViolation) + '\n'
        res = res + self.message
        return res

    def copy(self):
        """
        Return a deep copy.
        
        """
        return FitResult(bestFitParameters=self.bestFitParameters.copy(),
                         minChiSquare=self.minChiSquare, 
                         status=self.status,
                         success=self.success,
                         feasible=self.feasible,
                         message=self.message,
                         nEvaluations=self.nEvaluations,
                         constraintViolation=self.constraintViolation)

    def __lt__(self, r):
        if r is None:
            return True
        if not (hasattr(r, 'feasible') and hasattr(r, 'minChiSquare') and \
                hasattr(r, 'constraintViolation')):
            return NotImplemented
        if self.feasible and not r.feasible:
            return True
        elif self.feasible and r.feasible:
            if self.minChiSquare != r.minChiSquare:
                return self.minChiSquare < r.minChiSquare
            else:
                self.constraintViolation < r.constraintViolation
        elif not self.feasible and not r.feasible:
            if self.constraintViolation != r.constraintViolation:
                return self.constraintViolation < r.constraintViolation
            else:
                return self.minChiSquare < r.minChiSquare                
        else:
            return False

    def __eq__(self, r):
        if r is None:
            return False
        if not (hasattr(r, 'feasible') and hasattr(r, 'minChiSquare') and \
                hasattr(r, 'constraintViolation')):
            return NotImplemented
        return self.feasible == r.feasible and \
            self.minChiSquare == r.minChiSquare and \
            self.constraintViolation == r.constraintViolation


class _ObjectiveFunction:
    """
    Provides all information for constraned optimization.

    Args:
      model (Model): Model object describing the chi-square function to minimize
        and the constraints on the parameters.

    Kwargs:
      obs (dict or None): Dictionary with the 'measured' values of the
        observables. All keys returned by ``model.getObservableNames()``
        should be present in `obs`. Defaults to ``None``, in which case
        ``model.obs`` is used.
      origin (dict): Starting point of the minimization. Keys must be the
        elements of ``model.getFreeParameters()``. Values must be floats.
        Defaults to a dict with all values equal zero.
      basis (list of dicts): Defines a basis in parameter space. The length of
        the list must be equal to the number of free parameters (i.e.
        ``len(model.getFreeParameters())``). The keys of each dictionary
        must be the elements of ``model.getFreeParameters()``. Minimizations
        will be carried out in this basis. Defaults to a diagonal 'matrix' where
        each diagonal entry is the parameter's ``scale`` attribute.
    
    Attributes:
      npar (int): Number of free parameters.
      f (callable): Objective function. Takes a single numpy.ndarray as
        argument. The elements of this array identify the point in parameter space
        in the basis `basis` relative to the point specified by `origin`.
      lb, ub (list): List of lower/upper bounds on the parameters (in the basis
        `basis`). ``None`` indicates no lower/upper bound.
      A, b (numpy.ndarray): `A` is a matrix and `b` a vector. Together they
        specify linear constraints of the form ``A.x >= b`` (where the
        inequality holds element-wise).
      c, h (callable): Non-linear inequality and equality constraints. Both
        functions take the same argument `x` as `f` and return a list of
        floats. They describe the constraints ``c(x) >= 0, h(x) == 0``, where
        the inequality or equality holds for each element of the returned list.

    """

    def __init__(self, model, obs=None, origin=None, basis=None):
        self.f = None
        self.c = []
        self.h = []

        self._model = model
        self._obs = model.obs if obs is None else obs
        self._pkeys = model.getFreeParameters()
        self._fixedpar = model.getFixedParameters()
        self._ckeys = [k for k, c in model.con.iteritems() if c.active]
        self.neqcons = len([k for k in self._ckeys if model.con[k].equality])
        self.nieqcons = len(self._ckeys) - self.neqcons

        self.npar = len(self._pkeys)
        pind = dict((self._pkeys[i], i) for i in range(len(self._pkeys)))
        if origin is None:
            self._p0 = numpy.zeros(self.npar)
        else:
            self._p0 = numpy.array([origin.get(k, 0.0) for k in self._pkeys])


        self._x = numpy.array([])
        self._q = {}
        self._B = numpy.zeros((self.npar,self.npar))
        if basis is None:
            for i in range(self.npar):
                self._B[i,i] = model.par[self._pkeys[i]].scale
            self.A = []
            self.b = []
            self.lb = []
            self.ub = []
            for i in range(self.npar):
                if model.par[self._pkeys[i]].lowerBound is None:
                    self.lb.append(None)
                else:
                    self.lb.append((model.par[self._pkeys[i]].lowerBound - 
                                     self._p0[i])/self._B[i,i])
                if model.par[self._pkeys[i]].upperBound is None:
                    self.ub.append(None)
                else:
                    self.ub.append((model.par[self._pkeys[i]].upperBound - 
                                     self._p0[i])/self._B[i,i])
        else:
            for i in range(self.npar):
                for j in range(self.npar):
                    self._B[i,j] = basis[j].get(self._pkeys[i], 0.0)
            self.lb = [None]*self.npar
            self.ub = [None]*self.npar
            self.A = []
            self.b = []
            for i in range(self.npar):
                k = self._pkeys[i]
                if model.par[k].lowerBound is not None:
                    self.A.append(list(self._B[i]/model.par[k].scale))
                    self.b.append((model.par[k].lowerBound - self._p0[i])/
                                  model.par[k].scale)
                if model.par[k].upperBound is not None:
                    self.A.append(list(-self._B[i]/model.par[k].scale))
                    self.b.append((self._p0[i] - model.par[k].upperBound)/
                                  model.par[k].scale)

        self.A = numpy.array(self.A)
        self.b = numpy.array(self.b)
        self._Binv = numpy.linalg.inv(self._B)
        
    def getModelParameters(self, x):
        keys = self._pkeys
        pvec = numpy.dot(self._B, x) + self._p0
        p = dict((keys[i], pvec[i]) for i in range(len(keys)))
        p.update(self._fixedpar)
        return p 

    def getScaledParameters(self, p):
        pvec = numpy.array(p[k] for k in self._pkeys)
        return numpy.dot(self._Binv, pvec - self._p0)

    def getConstraintViolation(self, x):
        self.update(x)
        return sum(val**2 for val in self.h) + \
            sum(val**2 for val in self.c if val < 0.0)

    def update(self, x):
        if numpy.array_equal(x, self._x):
            return
        self._x = numpy.copy(x)
        p = self.getModelParameters(x)

        q = self._model.getDependentQuantities(p)
        con = self._model.getConstraints(q)
        self.f = self._model.getChiSquare(q, obs=self._obs)
        self.c = []
        self.h = []
        for k in self._ckeys:
            if self._model.con[k].equality:
                self.h.append(con[k])
            else:
                self.c.append(con[k])
        self.c = numpy.array(self.c)
        self.h = numpy.array(self.h)        


def _fit_scipy_cobyla(obj, rhobeg=1.0, acc=1e-5, maxiter=1000,
                      disp=False, iprint=1):
    def calcfc(x, con):
        obj.update(x)
        ctemp = list(obj.c)
        ctemp.extend(obj.h)
        ctemp.extend(-hi for hi in obj.h)
        if len(obj.A) > 0:
            ctemp.extend(numpy.dot(obj.A, x) - obj.b)
        ctemp.extend(obj.ub[i] - x[i] for i in range(len(obj.ub)) \
                         if obj.ub[i] is not None)
        ctemp.extend(x[i] - obj.lb[i] for i in range(len(obj.lb)) \
                         if obj.lb[i] is not None)
        for i in range(len(con)):
            con[i] = ctemp[i]
        return obj.f

    if not disp:
        iprint=0
    x0 = numpy.zeros(obj.npar)
    obj.update(x0)
    m = len(obj.c) + len(obj.h) + len(obj.A) + \
        len([b for b in obj.lb if b is not None]) + \
        len([b for b in obj.ub if b is not None])
    info = numpy.zeros(4, numpy.float64)
    xopt, info = _cobyla.minimize(calcfc, m=m, x=x0, rhobeg=rhobeg,
                                  rhoend=acc, iprint=iprint, maxfun=maxiter,
                                  dinfo=info)

    return FitResult(bestFitParameters=obj.getModelParameters(xopt),
                     minChiSquare=info[2],
                     status=int(info[0]),
                     success=info[0] == 1,
                     message={1: 'Optimization terminated successfully.',
                              2: 'Maximum number of function evaluations has '
                              'been exceeded.',
                              3: 'Rounding errors are becoming damaging in '
                              'COBYLA subroutine.'
                              }.get(info[0], 'Unknown exit status.'),
                     nEvaluations=int(info[1]),
                     constraintViolation=obj.getConstraintViolation(xopt))


def _fit_scipy_slsqp(obj, acc=1e-5, maxiter=1000, epsilon=1e-8,
                     disp=False, iprint=1):
    def func(x):
        obj.update(x)
        return obj.f

    def ieqcons(x):
        obj.update(x)
        c = obj.c
        if len(obj.A) > 0:
            return numpy.concatenate((obj.c, numpy.dot(obj.A, x) - obj.b))
        else:
            return obj.c

    def eqcons(x):
        obj.update(x)
        return obj.h

    lb = numpy.array([(-numpy.inf if b is None else b) for b in obj.lb])
    ub = numpy.array([(+numpy.inf if b is None else b) for b in obj.ub])

    x0 = numpy.zeros(obj.npar)
    out, fx, its, imode, smode = \
        fmin_slsqp(func, x0, f_eqcons=eqcons, f_ieqcons=ieqcons,
                   bounds=zip(lb,ub),
                   iter=maxiter, acc=acc, iprint=iprint, disp=disp,
                   full_output=True, epsilon=epsilon)

    return FitResult(bestFitParameters=obj.getModelParameters(out),
                     minChiSquare=fx,
                     status=imode,
                     success=imode == 0,
                     message=smode,
                     nEvaluations=its,
                     constraintViolation=obj.getConstraintViolation(out))


# def _fit_nlopt_slsqp(obj, acc=1e-5, maxiter=1000, ctol=1e-6, epsilon=1e-8,
#                      disp=False, iprint=1):
#     def _func(x):
#         obj.update(x)
#         return obj.f

#     dfunc = numdifftools.Gradient(_func, step_nom=[epsilon]*obj.npar)

#     def func(x, g):
#         result = _func(x)
#         if g.size > 0:
#             g[:] = dfunc(x)
#         return result

#     def _ieqcons(x):
#         obj.update(x)
#         if len(obj.A) > 0:
#             return numpy.concatenate((-obj.c, -numpy.dot(obj.A, x) + obj.b))
#         else:
#             return -obj.c

#     dieqcons = numdifftools.Jacobian(_ieqcons, step_nom=[epsilon]*obj.npar)

#     def ieqcons(result, x, g):
#         result[:] = _ieqcons(x)
#         if g.size > 0:
#             g[:,:] = dieqcons(x)

#     def _eqcons(x):
#         obj.update(x)
#         return obj.h

#     deqcons = numdifftools.Jacobian(_eqcons, step_nom=[epsilon]*obj.npar)

#     def eqcons(result, x, g):
#         result[:] = _eqcons(x)
#         if g.size > 0:
#             g[:,:] = deqcons(x)

#     lb = numpy.array([(-numpy.inf if b is None else b) for b in obj.lb])
#     ub = numpy.array([(+numpy.inf if b is None else b) for b in obj.ub])

#     x0 = numpy.zeros(obj.npar)

#     opt = nlopt.opt(nlopt.LD_SLSQP, obj.npar)
#     opt.set_min_objective(func)
#     opt.set_lower_bounds(lb)
#     opt.set_upper_bounds(ub)
#     opt.add_inequality_mconstraint(ieqcons, [ctol]*(obj.nieqcons+len(obj.b)))
#     opt.add_equality_mconstraint(eqcons, [ctol]*obj.neqcons)
#     opt.set_ftol_abs(acc)
#     opt.set_maxeval(maxiter)

#     try:
#         xopt = opt.optimize(x0)
#         status = opt.last_optimize_result()
#     except nlopt.RoundoffLimited as e:
#         xopt = obj._x
#         status = nlopt.ROUNDOFF_LIMITED

#     fmin = _func(xopt)
#     if status == nlopt.MAXEVAL_REACHED:
#         message = 'Maximum number of iterations exceeded.'
#     elif status == nlopt.ROUNDOFF_LIMITED:
#         message = 'Optimization limited by rounding errors.'
#     elif status != nlopt.SUCCESS:
#         message = 'Generic failure.'
#     else:
#         message = 'Optimization terminated successfully.'

#     return FitResult(bestFitParameters=obj.getModelParameters(xopt),
#                      minChiSquare=fmin,
#                      status=status,
#                      success=status == nlopt.SUCCESS,
#                      message=message,
#                      nEvaluations=-1,
#                      constraintViolation=obj.getConstraintViolation(xopt))


# def _fit_openopt_ralg(obj, acc=1e-5, maxiter=1000, epsilon=1e-8,
#                       disp=False, iprint=1):
#     if not disp:
#         iprint = -1

#     def f(x):
#         obj.update(x)
#         return obj.f

#     def c(x):
#         obj.update(x)
#         return -obj.c

#     def h(x):
#         obj.update(x)
#         return obj.h

#     lb = numpy.array([(-numpy.inf if b is None else b) for b in obj.lb])
#     ub = numpy.array([(+numpy.inf if b is None else b) for b in obj.ub])
#     A = -obj.A
#     b = -obj.b
#     x0 = numpy.zeros(obj.npar)

#     p = NLP(f=f, x0=x0, c=c, h=h, lb=lb, ub=ub, A=A, b=b,
#             iprint=iprint, maxIter=maxiter, maxFunEvals=obj.npar*maxiter,
#             ftol=acc, xtol=acc, gtol=acc)
#     r = p.solve('ralg')

#     if r.stopcase == 1:
#         r.msg = "Minimization terminated successfully."
#     return FitResult(bestFitParameters=obj.getModelParameters(r.xf),
#                      minChiSquare=r.ff,
#                      status=r.istop,
#                      success=r.stopcase == 1,
#                      message=r.msg,
#                      constraintViolation=obj.getConstraintViolation(r.xf),
#                      nEvaluations=r.evals['iter'])
        

def fit(model, obs=None, start=None, tries=None, basis=None, maxcv=1e-4,
        autoscale=False, epsilon=1e-8, maxcond=1e6, solver='slsqp',
        raiseFitError=True, solvercfg={}):
    """Do a local minimization of a model's chi-square function.

    Args:
      model (Model instance): The model to be fitted.
      obs (dict or None, optional): Dictionary with the 'measured' values of the
        observables. All keys returned by ``model.getObservableNames()``
        should be present in `obs`. Defaults to ``None``, in which case
        ``model.obs`` is used.
      start (dict or list of dicts or None, optional): Starting values of the
        parameters.  If `start` is a list or tuple of dicts the points in the
        list are tried in sequence until `tries` minimizations have been
        successful. The result of the successful minimization with the smallest
        chi-square is returned. Defaults to ``None``, in which case
        ``model.getDefaultParameters()`` is used.
      tries (int or None, optional): The number of successful minimizations to
        collect before aborting the iteration over the provided starting points.
        Defaults to ``None``, in which case all provided starting points are
        tried.
      basis (list of dicts or None, optional): This argument specifies a linear
        transformation in parameter space. The list must have as many entries
        as there are free parameters in the fit. Each dict specifies a
        vector in parameter space which is used as a basis vector. Defaults
        to ``None``, in which case the basis is constructed from the
        ``scale`` attributes of the model parameters.
      maxcv (float, optional): maximal value of the constraint violation (sum of
        squares of all violated constraint functions) for which the optimization
        is still considered successful. Defaults to ``1e-4``.
      autoscale (bool, optional): If ``True``, a basis for the optimization is
        constructed from the gradients of the observables. If ``False``, the
        basis is constructed from the ``scale`` attributes of the model
        parameters. Defaults to ``False``.
      epsilon (float, optional): Small number used to estimate the derivatives of
        observables with respect to parameters. This is needed for
        constructing a basis in parameter space. Irrelevant if
        ``autoscale=False``. Defaults to ``1e-8``.
      maxcond (float, optional): Used for re-scaling the optimization problem.
        The ratio of the length of the longest basis vector over the length
        of the shortest basis vector can at most be ``sqrt(maxcond)``. Defaults
        to ``1e6``. Irrelevant if ``autoscale=False``.
      solver (str, optional): Identifies the optimization algorithm to use.
        Currently supported values are:

          ``'cobyla'``:
            Uses scipy's fmin_cobyla function
          ``'slsqp'``:
            Uses scipy's fmin_slsqp function (default)

      raiseFitError (bool, optional): Whether or not to raise a FitError
        exception when the model raises a ParameterError exception. Defaults to
        ``True``.
      solvercfg (dict, optional): Keyword arguments that are passed to the
        solver. Common options are:

          acc (float)
            Accuracy of the minimization.
          maxiter (int)
            Maximum number of iterations.
          disp (bool)
            Whether to print information about the minimization.
          iprint (int)
            How much information to print.

        Defaults to ``{}``.
        
    Returns: 
      An instance of :py:class:`~myFitter.FitResult` with the results of the fit.

    """
    if not model.getFreeParameters():
        q = model.getDependentQuantities(model.getFixedParameters())
        chisq = model.getChiSquare(q, obs=obs)
        cv = model.getConstraintViolation(q)
        return FitResult(bestFitParameters=model.getFixedParameters(),
                         minChiSquare=chisq,
                         status=0,
                         success=True,
                         feasible=(cv <= maxcv),
                         message='No parameters to fit.',
                         nEvaluations=1,
                         constraintViolation=cv)

    if start is None:
        start = model.getDefaultParameters()
    if not (isinstance(start, list) or isinstance(start, tuple)):
        start = [start]
    if tries is None:
        tries = len(start)

    res = None
    fixedpar = model.getFixedParameters()
    defaultpar = model.getDefaultParameters()
    fixedbasis = (not autoscale) or (basis is not None)
    nsuccess = 0
    for orig in start:
        if nsuccess >= tries:
            break
        origin = defaultpar.copy()
        origin.update(orig)
        origin.update(fixedpar)
        if not fixedbasis:
            metric = model.getQuadraticForm(
                model.getDependentQuantities(origin))
            basis = _get_basis(model, obs, origin, metric, epsilon, maxcond)
        try:
            if solver == 'cobyla':
                r = _fit_scipy_cobyla(
                    _ObjectiveFunction(model, obs=obs, origin=origin,
                                       basis=basis), **solvercfg)
            elif solver == 'slsqp':
                r = _fit_scipy_slsqp(
                    _ObjectiveFunction(model, obs=obs, origin=origin,
                                       basis=basis), **solvercfg)
            # elif solver == 'nlopt_slsqp':
            #     r = _fit_nlopt_slsqp(
            #         _ObjectiveFunction(model, obs=obs, origin=origin,
            #                            basis=basis), **kwargs)
            # elif solver == 'ralg':
            #     r = _fit_openopt_ralg(
            #         _ObjectiveFunction(model, obs=obs, origin=origin,
            #                            basis=basis), **kwargs)
            else:
                raise ValueError('Unknown solver requested')

        except ParameterError as e:
            if raiseFitError:
                kwargs.update({'obs'           : obs,
                               'start'         : start,
                               'basis'         : basis,
                               'maxcv'         : maxcv,
                               'autoscale'     : autoscale,
                               'epsilon'       : epsilon,
                               'maxcond'       : maxcond,
                               'solver'        : solver,
                               'raiseFitError' : raiseFitError})
                raise FitError(e.message, model, kwargs)
            else:
                q = model.getDependentQuantities(origin)
                r = FitResult(bestFitParameters=origin,
                              minChiSquare=model.getChiSquare(q, obs),
                              status=-100,
                              success=False,
                              message='Unknown solver error.',
                              nEvaluations=-1,
                              constraintViolation=\
                                  model.getConstraintViolation(q))

        r.feasible = (r.constraintViolation <= maxcv)
        if r.feasible and r.success:
            nsuccess = nsuccess + 1
        if res is None:
            res = r
        else:
            res = min(res, r)

    return res


def _get_basis(model, obs, origin, metric, epsilon, maxcond):
    par = model.getFreeParameters()
    npar = len(par)
    if npar == 0:
        return None
    obsnames = model.getObservableNames()
    connames = model.con.keys()
    nobs = len(obsnames)
    q0 = model.getDependentQuantities(origin)
    f0 = model.getChiSquare(q0, obs)
    x0 = _dict_to_vector(model.getCentralValues(q0), obsnames)
    c0 = _dict_to_vector(model.getConstraints(q0), connames)
    metric = _dict_to_matrix(metric, obsnames)
    p1 = origin.copy()

    dxdp = []  # first index: parameter, second index: observable
    dcdp = []  # first index: parameter, second index: constraint
    gradf = []
    for p in par:
        delta = model.par[p].scale*epsilon
        p1[p] = origin[p] + delta
        q = model.getDependentQuantities(p1)
        f1 = model.getChiSquare(q, obs)
        x1 = _dict_to_vector(model.getCentralValues(q), obsnames)
        c1 = _dict_to_vector(model.getConstraints(q), connames)
        dxdp.append((x1 - x0)/epsilon)
        dcdp.append((c1 - c0)/epsilon)
        gradf.append((f1-f0)/delta)
        p1[p] = origin[p]
    dxdp = numpy.array(dxdp)
    dcdp = numpy.array(dcdp)
    gradf = numpy.array(gradf)
    
    pmetric = numpy.dot(dxdp, numpy.dot(metric, numpy.transpose(dxdp))) + \
        numpy.dot(dcdp, numpy.transpose(dcdp))
    for i, p in enumerate(par):
        if model.par[p].lowerBound is not None or \
                model.par[p].upperBound is not None:
            pmetric[i,i] += 1.0

    try:
        lam, B = numpy.linalg.eigh(pmetric)
    except numpy.linalg.LinAlgError:
        return None
    for i in range(npar):
        B[i,:] *= model.par[par[i]].scale
    lam = abs(lam)
    lmin = max(lam)/maxcond
    lam = [max(l, lmin) for l in lam]
    for i in range(npar):
        B[:,i] /= sqrt(lam[i])
    scale = numpy.linalg.norm(numpy.dot(numpy.transpose(B), gradf))
    if scale <= 0.0:
        return None
    B *= 1.0/scale

    return [_vector_to_dict(B[:,i], par) for i in range(npar)]
    

def pvalueFromZvalue(z):
    """Computes the p-value from the Z-value.

    Note:
      The Z-value is often referred to as the 'number of standard deviations' or
      'number of sigmas'. myFitter uses two-sided Z-values which are related to
      the p-value by  p = 1 - Erf(Z/sqrt(2))

    Args:
      z (float): The Z-value.

    Returns:
      float: The p-value corresponding to the Z-value `z`.

    """
    return 1.0-erf(z/sqrt(2.0))

def zvalueFromPvalue(p):
    """Computes the Z-value from the p-value.

    Note:
      The Z-value is often referred to as the 'number of standard deviations' or
      'number of sigmas'. myFitter uses two-sided Z-values which are related to
      the p-value by  p = 1 - Erf(Z/sqrt(2))

    Args:
      p (float): The p-value.

    Returns:
      float: The Z-value corresponding to the p-value `p`.

    """
    return sqrt(2.0)*erfinv(1.0 - p)

def pvalueFromChisq(chisq, ndof):
    """Computes the p-value from the chi-square value.

    Args:
      chisq (float): The chi-square value.
      ndof (int): The number of degrees of freedom.

    Returns:
      float: The p-value corresponding to the chi-square value `chisq`, using
        Wilks' theorem with `ndof` degrees of freedom.

    """
    return 1.0 - gammainc(0.5*ndof, 0.5*chisq)

def chisqFromPvalue(p, ndof):
    """Computes the chi-square value from the p-value.

    Args:
      p (float): The p-value.
      ndof (int): The number of degrees of freedom.

    Returns:
      float: The chi-square value corresponding to the p-value `p`, using Wilks'
        theorem with `ndof` degrees of freedom.

    """
    return 2.0*gammaincinv(0.5*ndof, 1.0 - p)

def zvalueFromChisq(chisq, ndof):
    """Computes the Z-value from the chi-square value.

    Note:
      The Z-value is often referred to as the 'number of standard deviations' or
      'number of sigmas'. myFitter uses two-sided Z-values which are related to
      the p-value by  p = 1 - Erf(Z/sqrt(2))

    Args:
      chisq (float): The chi-square value.
      ndof (int): The number of degrees of freedom.

    Returns:
      float: The Z-value corresponding to the chi-square value `chisq`, using
        Wilks' theorem with `ndof` degrees of freedom.

    """
    return zvalueFromPvalue(pvalueFromChisq(chisq, ndof))

def chisqFromZvalue(z, ndof):
    """Computes the chi-square value from the Z-value.

    Note:
      The Z-value is often referred to as the 'number of standard deviations' or
      'number of sigmas'. myFitter uses two-sided Z-values which are related to
      the p-value by  p = 1 - Erf(Z/sqrt(2))

    Args:
      z (float): The Z-value.
      ndof (int): The number of degrees of freedom.

    Returns:
      float: The chi-square value corresponding to the Z-value `z`, using Wilks'
        theorem with `ndof` degrees of freedom.

    """
    return chisqFromPvalue(pvalueFromZvalue(z), ndof)

