from scipy.interpolate import InterpolatedUnivariateSpline
import numpy
import collections
import bisect

class InterpolationError(Exception):
    pass

def _interp2(x0, x1, y0, y1, v0, x):
    dx = x1 - x0
    dy = y1 - y0
    r  = (x - x0)/dx
    return (dy - v0*dx)*r**2 + v0*dx*r + y0

def _dinterp2(x0, x1, y0, y1, v0):
    dx = x1 - x0
    dy = y1 - y0
    return 2.0*(dy - v0*dx)/dx + v0

def _interp3(x0, x1, y0, y1, v0, v1, x):
    dx = x1 - x0
    dy = y1 - y0
    r  = (x - x0)/dx
    a  = (v1 + v0)*dx - 2.0*dy
    b  = 3.0*dy - (v1 + 2.0*v0)*dx
    return a*r**3 + b*r**2 + v0*dx*r + y0

class Interpolator1D:
    def __init__(self, file=None, filename=None, data=None, xcol=0, ycol=1):
        if filename is not None:
            file = open(filename, 'r')
        if file is not None:
            data = []
            for line in file:
                if line.strip()[0] == '#':
                    continue
                cols = line.split()
                data.append([float(col) for col in cols])
            file.close()
        if data is not None:
            data = sorted(data, key=lambda r: r[xcol])
            self._x = [r[xcol] for r in data]
            if hasattr(ycol, '__iter__'):
                self._y = [numpy.array([r[i] for i in ycol]) for r in data]
            else:
                self._y = [r[ycol] for r in data]
        else:
            raise InterpolationError('No data to interpolate.')
        if len(self._x) < 2:
            raise InterpolationError('Need at least two data points.')


    def __call__(self, x):
        xv = self._x
        yv = self._y
        if len(xv) <= 2:
            return (yv[1] - yv[0])/(xv[1]-xv[0])*(x - xv[0]) + yv[0]
        i = bisect.bisect_left(self._x, x)
        if i == 0:
            x0 = xv[0]
            y0 = yv[0]
            v0 = _dinterp2(xv[1], x0, yv[1], y0,
                           (yv[2]-y0)/(xv[2]-x0))
            return v0*(x-x0) + y0
        elif i == len(xv):
            x0 = xv[-1]
            y0 = yv[-1]
            v0 = _dinterp2(xv[-2], x0, yv[-2], y0,
                                (y0 - yv[-3])/(x0 - xv[-3]))
            return v0*(x-x0) + y0
        elif i == 1:
            return _interp2(xv[1], xv[0], yv[1], yv[0],
                            (yv[2] - yv[0])/(xv[2] - xv[0]), x)
        elif i == len(xv)-1:
            return _interp2(xv[-2], xv[-1], yv[-2], yv[-1],
                            (yv[-1] - yv[-3])/(xv[-1] - xv[-3]), x)
        else:
            return _interp3(xv[i-1], xv[i], yv[i-1], yv[i],
                            (yv[i]-yv[i-2])/(xv[i]-xv[i-2]),
                            (yv[i+1]-yv[i-1])/(xv[i+1]-xv[i-1]), x)

