import numpy
import scipy
import vegas
from model import _dict_to_vector, _vector_to_dict, _dict_to_matrix
from fitter import MetricError
from histograms import Histogram1D, Histogram2D, GVar, IntegrationResult, \
    Integrator
from math import sqrt, exp, log, pi
from scipy.special import erf, erfinv
import sys


def _ortho(M, B=None, tol=1e-6):
    if B is not None and M.shape[0] != B.shape[0]:
        raise ValueError('M and B must have the same number of rows.')
    if B is not None and B.shape[1] >= B.shape[0]:
        return B.copy()
    dim, nM = M.shape
    if B is not None:
        nB = B.shape[1]
        A = M - numpy.dot(B, numpy.dot(B.T, M))
    else:
        nB = 0
        A = M
    U, s, VT = scipy.linalg.svd(A)
    # tol = max(dim,nM) * numpy.amax(s) * epsilon
    n = min(numpy.sum(s > tol, dtype=int), dim-nB)
    return U[:,:n]


def _xdxdp(model, obsnames, par={}, epsilon=1e-6, minrange=2.5):
    parnames = []
    for n in model.getFreeParameters():
        p = model.par[n]
        if p.lowerBound is None or p.upperBound is None or \
           abs(p.upperBound-p.lowerBound) > p.scale*minrange:
            parnames.append(n)
    par0 = model.getDefaultParameters()
    par0.update(par)
    pvec0 = _dict_to_vector(par0, parnames)
    scales = numpy.array([model.par[p].scale for p in parnames])
    q = model.getDependentQuantities(par0)
    xvec0 = _dict_to_vector(model.getCentralValues(q), obsnames)

    dxdp = numpy.empty((len(obsnames), len(pvec0)), dtype=float)
    pvec = pvec0.copy()
    for i in range(len(pvec0)):
        pvec[i] = pvec[i] + epsilon*scales[i]
        par0.update(_vector_to_dict(pvec, parnames))
        q = model.getDependentQuantities(par0)
        xvec = _dict_to_vector(model.getCentralValues(q), obsnames)
        dxdp[:,i] = (xvec - xvec0)/(epsilon*scales[i])
        pvec[i] = pvec0[i]
    return xvec0, dxdp


class _Integrand:
    def __init__(self, f, integrator):
        self._i = integrator
        self._f = f
        self._vectogauss = \
            numpy.vectorize(lambda p: sqrt(2.0)*erfinv(2.0*p-1.0))

    def _togauss(self, x):
        return self._vectogauss(x) if x.size > 0 else x

    def _getlam(self, x, dsq=None):
        if dsq is None:
            lam = self._togauss(x)
            chisq0 = len(lam)*log(2.0*pi) + numpy.vdot(lam,lam)
            return lam, chisq0
        else:
            f = self._i._ifrac
            alpha = self._i._ipow
            ndim = len(x)-1
            p0=scipy.special.gammainc(0.5*ndim,0.5*dsq)

            if x[0] < f:
                rsq = dsq*(x[0]/f)**(2.0/(ndim+alpha))
                chisq0 = \
                    +2.0*log(2.0)+ndim*log(pi) \
                    +(ndim+alpha)*log(dsq) \
                    -2.0*log(f*(ndim+alpha)) \
                    -2.0*scipy.special.gammaln(0.5*ndim) \
                    -alpha*log(rsq)
            else:
                p = (x[0]-f)/(1.0-f)
                rsq = 2.0*scipy.special.gammaincinv(0.5*ndim,(1.0-p0)*p + p0)
                if p0 == 1.0:
                    chisq0 = -numpy.inf
                else:
                    chisq0 = ndim*log(2.0*pi) + 2.0*log((1.0-p0)/(1.0-f)) + rsq
            lam = self._togauss(x[1:])
            lsq = numpy.vdot(lam, lam)
            if lsq > 0.0:
                lam = lam*sqrt(rsq/lsq)
            return lam, chisq0
        
    def __call__(self, x):
        if not self._i._smart:
            lam, chisq0 = self._getlam(x)
        else:
            lam1, chisq1 = self._getlam(x[0:self._i._n0])
            if self._i._lam0 is not None:
                dsq = self._i._dchisq + numpy.linalg.norm(lam1-self._i._lam0)**2
            else:
                dsq = self._i._dchisq
            lam2, chisq2 = self._getlam(x[self._i._n0:self._i._n1+1], dsq)
            lam3, chisq3 = self._getlam(x[self._i._n1+1:])
            lam = numpy.concatenate((lam1, lam2, lam3))
            chisq0 = chisq1 + chisq2 + chisq3

        obs    = _vector_to_dict(self._i._x + numpy.dot(self._i._B, lam),
                                 self._i._obsnames)
        chisq  = self._i._model.getChiSquare(self._i._q, obs=obs)
        weight = self._i._jacobian*exp(0.5*(chisq0-chisq))

        return numpy.asarray(self._f(obs))*weight


class ToySimulator(Integrator):
    """Class for running toy simulations and computing p-values.

    Note:
      ToySimulator is a subclass of :py:class:`~myFitter.Integrator`.  It
      can use a "smart" sampling method which is taylored to p-value
      computations in (nested or non-nested) LRTs. To do this it requires
      additional information about the alternative hypothesis and the Delta
      chi-square. If you do not provide this information ToySimulator uses
      standard sampling (i.e. generates points distributed according to the PDF
      associated with your null hypothesis).

    Args:
      model0 (Model): Model instance which defines the PDF.
      par0 (dict, optional): Parameter values for the model. Model defaults are 
        used for unspecified parameters. Defaults to ``{}``.
      model1 (Model or None, optional): Model instance which describes the
        alternative hypothesis. Only required if you want to use smart sampling.
        Defaults to ``None``.
      par1 (dict or None, optional): Parameter values for `model1`. Model
        defaults are used for unspecified parameters. Defaults to ``{}``.
      dchisq (float, optional): Delta chi-square value for the LRT. Only
        relevant if you want to use smart sampling. Defaults to 0.
      smart (bool, optional): Whether to use smart sampling. Defaults to
        ``True``.
      ifrac (float, optional): Fraction of sample points to put in "inner
        region". Only relevant if you use smart sampling. Defaults to 0.25.
      ipow (float, optional): Power law to use for sampling the "inner region".
        Only relevant if you use smart sampling. Defaults to 2.
      minrange (float, optional): When using smart sampling, parameters which
        only float in a very small range should be considered as fixed for the
        purpose of constructing the smart sampling density. If the range of a
        parameter is smaller than `minrange` times its
        :py:attr:`~Parameter.scale` attribute it is considered as fixed for this
        purpose. Defaults to 2.5.
      epsilon (float, optional): Small parameter to compute numerical
        derivatives. Only relevant if you use smart sampling. Defaults to
        ``1e-6``.
      tol (float, optional): Tolerance to use for orthogonalising vectors. Only
        relevant if you use smart sampling. Defaults to ``1e-6``.
      **kwargs (optional): Additional keyword arguments are passed down to the
        constructor of ``vegas.Integrator``. The most important keywords are:

          `nitn`
             The number of iterations.
          `neval`
             The number of function evaluations per iteration.

    """
    def __init__(self, model0, par0={}, model1=None, par1=None, dchisq=0.0,
                 smart=True, ifrac=0.25, ipow=2.0, minrange=2.5, epsilon=1e-6,
                 tol=1e-6, **kwargs):
        if par1 is None:
            par1 = par0
        if dchisq <= 0.0:
            smart = False

        self._model    = model0
        self._dchisq   = dchisq
        self._smart    = smart
        self._ifrac    = ifrac
        self._ipow     = ipow
        self._obsnames = model0.getObservableNames()
        self._nobs     = len(self._obsnames)
        self._q = model0.getDependentQuantities(par0)
        self._x = _dict_to_vector(model0.getCentralValues(self._q),
                                  self._obsnames)
        Q       = _dict_to_matrix(model0.getQuadraticForm(self._q),
                                  self._obsnames)

        # construct basis _B
        lam, self._B = numpy.linalg.eigh(Q)
        Binv = self._B.T.copy()
        self._jacobian = 1.0
        for i in range(self._nobs):
            if lam[i] <= 0.0:
                raise MetricError('Indefinite metric in observable space')
            s = 1.0/sqrt(lam[i])
            self._jacobian *= s
            self._B[:,i] *= s
            Binv[i,:] /= s

        # align with model hypersurfaces
        if not smart:
            self._n0 = 0
            self._n1 = 0
            self._lam0 = None
        elif model1 is None:
            self._n0 = 0
            self._n1 = self._nobs
            self._lam0 = None
        else:
            if set(self._obsnames) != set(model1.getObservableNames()):
                raise ValueError('model0 and model1 must have the same observable space.')
            x0, M0 = _xdxdp(model0, self._obsnames, par0,
                            epsilon=epsilon, minrange=minrange)
            x0 = numpy.dot(Binv, x0)
            M0 = numpy.dot(Binv, M0)

            x1, M1 = _xdxdp(model1, self._obsnames, par1,
                            epsilon=epsilon, minrange=minrange)
            x1 = numpy.dot(Binv, x1)
            M1 = numpy.dot(Binv, M1)

            H0 = _ortho(M0, tol=tol)
            H1 = _ortho(M1, tol=tol)
            nested = (_ortho(M0, H1, tol=tol).shape[1] == 0)
            if nested:
                H1 = _ortho(M1, H0, tol=tol)
                H2 = _ortho(numpy.eye(self._nobs),
                            numpy.concatenate((H0,H1), axis=1), tol=tol)
                self._n0 = 0
                self._n1 = H1.shape[1]
                V = numpy.concatenate((H1, H0, H2), axis=1)
                self._B = numpy.dot(self._B, V)
                self._lam0 = None
            else:
                H0 = _ortho(numpy.dot(numpy.dot(H0, H0.T), M1), tol=tol)
                H1 = _ortho(M1, H0, tol=tol)
                H2 = _ortho(numpy.eye(self._nobs),
                            numpy.concatenate((H0,H1), axis=1), tol=tol)
                V = numpy.concatenate((H1, H2, H0), axis=1)
                self._n0 = H1.shape[1]
                self._n1 = H1.shape[1]+H2.shape[1]
                self._B = numpy.dot(self._B, V)
                self._lam0 = numpy.dot(V.T, x1 - x0)[:self._n0]

        # initialize base class
        if not smart:
            Integrator.__init__(self, self._nobs*[[0.0, 1.0]], **kwargs)
        else:
            Integrator.__init__(self, (self._nobs+1)*[[0.0, 1.0]], **kwargs)
                
    def __call__(self, g, **kwargs):
        """Integrate a function on observable space.

        Args:
          g (callable): The function to integrate. `g` should take a
            dict of observable values as argument and may return a float or
            a one-dimensional numpy array of floats. This computes the integral

              .. math::
                 \int d^n\\vec{x} g(\\vec{x}) f(\\vec{x}, \\vec{\\omega}_0)

            where :math:`f` is the PDF of `model0` and :math:`\\vec\\omega_0`
            the parameter point `par0` (both specified in the constructor).
          **kwargs (optional): Additional keyword arguments are passed down to
            the ``__call__`` method of :py:class:`~myFitter.Integrator`. The
            most important keywords are:

              `nitn`
                 The number of iterations.
              `neval`
                 The number of function evaluations per iteration.
              `parallelize`
                 ``dict`` of parallelisation options.

        """
        integrand = _Integrand(g, self)
        return Integrator.__call__(self, integrand, **kwargs)

    def random(self, **kwargs):
        """Generate random points in observable space.

        Note:
          This function returns an iterator which generates points in observable
          space distributed according to a Gaussian approximation of the PDF of
          `model0` with parameters `par0` (both set in the constructor). If
          `model0` has only Gaussian observables the approximation is exact.

        Args:
          **kwargs (optional): Keyword arguments to be passed to 
            :py:class:`Integrator.random`. The most important one is `neval`,
            which defines the maximal number of points yielded by the iterator.

        Yields:
          obs (dict): The point in observable space.
          weight (float): The weight of the point. The iterator yields `neval`
            points and their weights sum to 1. For models with only Gaussian
            observables all weights will be the same.

        """
        for item in Integrator.random(self, **kwargs):
            x = item[0]
            w = item[1]
            lam = numpy.vectorize(lambda p: sqrt(2.0)*erfinv(2.0*p-1.0))(x)
            obs = _vector_to_dict(self._x + numpy.dot(self._B, lam),
                                  self._obsnames)
            chisq0 = self._nobs*log(2.0*pi) + numpy.vdot(lam, lam)
            chisq  = self._model.getChiSquare(self._q, obs=obs)
            weight = self._jacobian*exp(0.5*(chisq0-chisq))
            yield tuple([obs, w*weight]+list(item[2:]))

    def fill(self, g, neval, *hist, **kwargs):
        """Fill histograms with data.

        Args:
          g (callable): function to integrate.
          neval (int): number of evaluations.
          *hist: tuples describing the histograms to fill.
          **kwargs (optional): Allowed keyword arguments are the same as for
            the constructor and can be used to override settings made there.

        Note:
          The function `f` must be function on observable space. It should take
          a ``dict`` `x` representing the point in observable space as argument
          and return an array whose first element is the function value of the
          integrand. The other elements can be 'binned'. The `*hist` arguments
          must be of the form ``(j, h)`` where ``h`` is a
          :py:class:`~myFitter.Histogram1D` or :py:class:`~myFitter.Histogram2D`
          instance and `j` can be a callable object, a tuple of integers or an
          integer.

          If h is a :py:class:`~myFitter.Histogram1D` instance and `j` is an
          integer each bin of `h` with boundaries `a` and `b` will be filled
          with a GVar object holding the estimate of the integral of ``f(x)[0]``
          over the set defined by ``a <= f(x)[j] < b``. Alternatively `j` can be
          a tuple of one integer. If `j` is a callable object and `h` a
          :py:class:`~myFitter.Histogram1D` instance the integration domain for
          each bin is defined by ``a <= j(x, f(x)) < b``.

          If h is a :py:class:`~myFitter.Histogram2D` instance and `j` is a
          tuple of two integers each bin of `h` with x-boundaries `ax` and `bx`
          and y-boundaries `ay` and `by` will contain an estimate of the
          integral of ``f(x)[0]`` over the domain with ``ax <= f(x)[j[0]] < bx``
          and ``ay <= f(x)[j[1]] < by``. If `j` is a callable it should return a
          tuple of two numbers and the integration domain is defined by ``ax <=
          j(x, f(x))[0] < bx`` and ``ay <= j(x, f(x))[1] < by``.

        Returns:
          GVar: The value (and error) of the integral.

        """
        integrand = _Integrand(g, self)
        return Integrator.fill(self, integrand, neval, *hist, **kwargs)
        

class LRTIntegrand:
    """Integrand function for computing p-values in LRTs.

    Note:
      Pass instances of ``LRTIntegrand`` to
      :py:meth:`myFitter.ToySimulator.__call__` to compute the p-values
      numerically.

    Args:
      fitter0 (Fitter): :py:class:`Fitter` instance which fits the model
        representing the null hypothesis of the LRT.
      fitter1 (Fitter): :py:class:`Fitter` instance which fits the model
        representing the alternative hypothesis of the LRT.
      dchisq (float or None, optional): Observed value of the Delta chi-square.
        Defaults to ``None``.
      nested (bool, optional): Set to ``True`` when the models are nested, i.e.
        when ``fitter0.model`` was obtained from ``fitter1.model`` by fixing
        parameters or adding constraints. Defaults to ``False``.
      verbose (int, optional): Amount of information to print to stdout. Defaults
        to 0, in which case no information is printed.

    Attributes:
      fitter0 (Fitter): Same as `fitter0` argument.
      fitter1 (Fitter): Same as `fitter1` argument.
      dchisq (float or None): Same as `dchisq` argument.
      nested (bool): Same as `nested` argument.
      verbose (int): Same as `verbose` argument.
      nshots (int): Number of integrand evaluations (since last call to
        :py:meth:`myFitter.LRTIntegrand.reset`).
      nfail (int): Number of failed integrand evaluations (since last call to
        :py:meth:`myFitter.LRTIntegrand.reset`).
      
    """
    def __init__(self, fitter0, fitter1, dchisq=None, nested=False, verbose=0):
        self.fitter0   = fitter0
        self.fitter1   = fitter1
        self.dchisq    = dchisq
        self.nested    = nested
        self.verbose   = verbose
        self.nshots    = 0
        self.nfail     = 0
            
    def fit0(self, x):
        if self.verbose == 2:
            sys.stdout.write('\nshot {0}: fitting model 1...'.format(
                    self.nshots))
            sys.stdout.flush()
        result = self.fitter0.fit(obs=x)
        if self.verbose == 2:
            sys.stdout.write('done.\n'+str(result))
            sys.stdout.flush()
        return result

    def fit1(self, x, start=[]):
        if self.verbose == 2:
            sys.stdout.write('\nshot {0}: fitting model 2...'.format(
                    self.nshots))
            sys.stdout.flush()
        result = self.fitter1.fit(obs=x, start=start)
        if self.verbose == 2:
            sys.stdout.write('done.\n'+str(result))
            sys.stdout.flush()
        return result
            
    def __call__(self, x):
        """Evaluate the integrand.

        Args:
          x (dict): The point in observable space.

        Returns:
          numpy array of 2 floats: The second element is the Delta chi-square
            value for x. The first element is 0 if this value is smaller than the
            value of `dchisq` set in the constructor or if one of the fits
            failed to find a feasible solution. In all other cases the first
            element is 1.

        """
        self.nshots = self.nshots + 1
        if self.verbose == 1:
            sys.stdout.write('shot {0}...'.format(self.nshots))
            sys.stdout.flush()            
        r0 = self.fit0(x)
        if self.nested:
            r1 = self.fit1(x, start=[r0.bestFitParameters])
        else:
            r1 = self.fit1(x)
        dchisq = r0.minChiSquare - r1.minChiSquare
        weight = 1.0
        if not (r0.feasible and r1.feasible):
            weight = 0.0
        if self.dchisq is not None and dchisq < self.dchisq:
            weight = 0.0

        if not (r0.success and r0.feasible and r1.success and r1.feasible):
            if self.verbose == 1:
                sys.stdout.write('fail.\n')
                sys.stdout.flush()
            self.nfail = self.nfail + 1
        else:
            if self.verbose == 1:
                sys.stdout.write('dchisq={0:g}.\n'.format(dchisq))
                sys.stdout.flush()

        return numpy.array([weight, dchisq])

    def reset(self):
        self.nshots = 0
        self.nfail  = 0
        
