import unittest
from myFitter import *

def relDev(a, b):
    return 0.0 if a == b == 0.0 else 2.0*(a-b)/(abs(a)+abs(b))

class MyModel(Model):
    def __init__(self):
        Model.__init__(self)

        self.addParameter('p1')
        self.addParameter('p2')
        self.addParameter('p3', generator=lambda: random.gauss(0.0,1.0))

        self.addConstraint('p1sq+p2sq', '=', 1.0, name='circle')
        self.addConstraint('p1p3', '<', +1.0, name='hyper1')
        self.addConstraint('p1p3', '>', -0.5, name='hyper2')
        self.addConstraint('p1sq+p2sq', '=', 2.0,
                            active=False, name='circle_2')

        self.addCorrelatedGaussianCSC('o1o2o3',
            [('o1', -1.0, 0.1),
             ('o2',  2.0, 0.6),
             ('o3', -0.5, 0.3)],
            [('o2','o1', -0.5),
             ('o3','o1',  0.8),
             ('o3','o2', -0.2)])
        self.addGaussianCSC('o4', 0.0, 0.1)
        self.addGaussianCSC('o5', 0.1, 0.03)
        
        self.csc['o3_2'] = GaussianCSC('o3', 0.2, active=False)

    def getDependentQuantities(self, p):
        q = Model.getDependentQuantities(self, p)

        q['p1sq+p2sq'] = p['p1']**2 + p['p2']**2
        q['p1p3'] = p['p1']*p['p3']

        q['o1'] = p['p1'] + p['p2']
        q['o2'] = p['p1'] - p['p2']
        q['o3'] = p['p2'] + p['p3']
        q['o4'] = p['p1'] - p['p3']
        q['o5'] = p['p1']*p['p2']*p['p3']

        return q


class ModelTest(unittest.TestCase):
    def test_dependentQuantities(self):
        model = MyModel()
        p = {'p1' :  0.4,
             'p2' : -1.2,
             'p3' :  0.2}
        q = model.getDependentQuantities(p)
        c = model.getConstraints(q)

        self.assertAlmostEqual(relDev(q['o1'], -0.8), 0.0, places=13)
        self.assertAlmostEqual(relDev(q['o2'],  1.6), 0.0, places=13)
        self.assertAlmostEqual(relDev(q['o3'], -1.0), 0.0, places=13)
        self.assertAlmostEqual(relDev(q['o4'],  0.2), 0.0, places=13)
        self.assertAlmostEqual(relDev(q['o5'], -0.096), 0.0, places=13)

        self.assertAlmostEqual(relDev(abs(c['circle']), 0.6), 0.0, places=13)
        self.assertAlmostEqual(relDev(c['hyper1'], 0.92), 0.0, places=13)
        self.assertAlmostEqual(relDev(c['hyper2'], 0.58), 0.0, places=13)

    def test_chiSquare(self):
        model = MyModel()
        p = {'p1' :  0.4,
             'p2' : -1.2,
             'p3' :  0.2}
        q = model.getDependentQuantities(p)
        self.assertAlmostEqual\
            (relDev(model.getChiSquare(q), 75.61583508314913), 0.0,
             places=13)
        csq123 = 36.8739224865
        csq4   = 1.23270688042
        csq5   = 37.5092057162
        self.assertAlmostEqual(
            relDev(model.chisq(p, only=['o1o2o3', 'o4']), csq123+csq4),
            0.0, places=12)
        self.assertAlmostEqual(
            relDev(model.chisq(p, only=['o1o2o3', 'o4'], exclude=['o4']),
                   csq123),
            0.0, places=12)
        self.assertAlmostEqual(
            relDev(model.chisq(p, exclude=['o1o2o3', 'o4']), csq5),
            0.0, places=12)

    def test_errors(self):
        model = MyModel()
        p = {'p1' :  0.4,
             'p2' : -1.2,
             'p3' :  0.2}
        q = model.getDependentQuantities(p)

        self.assertRaises(
            ValueError,
            lambda: CorrelatedGaussianCSC(
                {'o4'  : 1.0,
                 'o5'  : 1.0},
                {('o4','o5') : 1.0}))

        del q['o2']
        self.assertRaises(KeyError, lambda: model.getChiSquare(q))

    def test_quadraticApprox(self):
        model = MyModel()
        q = {'o1':0.3, 'o2': 1.2, 'o3':-0.6, 'o4':-2.9, 'o5':0.7}
        x0 = model.getCentralValues(q)
        Q = model.getQuadraticForm(q)
        self.assertAlmostEqual(relDev(x0['o1'], 0.3), 0.0, places=13)
        self.assertAlmostEqual(relDev(x0['o2'], 1.2), 0.0, places=13)
        self.assertAlmostEqual(relDev(x0['o3'],-0.6), 0.0, places=13)
        self.assertAlmostEqual(relDev(x0['o4'],-2.9), 0.0, places=13)
        self.assertAlmostEqual(relDev(x0['o5'], 0.7), 0.0, places=13)
        Qc = [[417.391304347826, 24.6376811594203, -101.449275362319, 0., 0.], 
              [24.6376811594203, 4.34782608695652, -4.83091787439613, 0., 0.], 
              [-101.449275362319, -4.83091787439613, 36.231884057971, 0., 0.], 
              [0., 0., 0., 100., 0.],
              [0., 0., 0., 0., 1111.11111111111]]
        o = ['o1', 'o2', 'o3', 'o4', 'o5']
        for i in range(len(o)):
            for j in range(len(o)):
                self.assertAlmostEqual(relDev(Q.get((o[i],o[j]), 0.0),
                                              Qc[i][j]), 0.0, places=13)

    def test_copy(self):
        model1 = MyModel()
        model2 = model1.copy()

        model1.con['circle'].bound = 70.0
        del model1.obs['o5']
        model1.csc['o4'].err = 90.0
        model1.csc['o1o2o3'] = GaussianCSC('o1', 10.0)
        model1.par['p1'].scale = 40.0

        self.assertTrue(model2.con['circle'].bound == 1.0)
        self.assertTrue('o5' in model2.obs)
        self.assertTrue(model2.csc['o4'].err == 0.1)
        self.assertTrue(isinstance(model2.csc['o1o2o3'],
                                   CorrelatedGaussianCSC))
        self.assertTrue(model2.par['p1'].scale == 1.0)

        
if __name__ == "__main__":
    unittest.main()

